/* ---------------------------------------------------------- 
%   (C)1994,1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#ifndef _KLIC_G_BASIC_H_
#define _KLIC_G_BASIC_H_

#include <assert.h>
#include <klic/stdc.h>  /* CONCATENATE*, STRINGIFY */

/* external module/class reference */

#define G_USE_PREDICATE(name) \
extern const struct predicate name

/* type judgement. */
#define G_ISCONS(x) iscons(x)
#define G_ISFUNCTOR(x) isfunctor(x)
#define G_ISREF(x) isref(x)
#define G_ISINT(x) isint(x)
#define G_ISSYM(x) issym(x)
#define G_ISGOBJ(x) (isfunctor(x) && isref(functor_of(x)))

/* data manipulation macro. */
#define G_FUNCTOR_OF(s) functor_of(s)
#define G_ARG(s, k) arg(s, k)
#define G_CAR_OF(x) car_of(x)
#define G_CDR_OF(x) cdr_of(x)
#define G_DEREFONE(x) derefone(x)

/* convertion. C->KLIC. */
#define G_MAKEINT(n) makeint(n)
#define G_MAKEREF(x) makeref(x)
#define G_MAKECONS(x) makecons(x)
#define G_MAKEFUNCTOR(x) makefunctor(x)

/* convertion. KLIC->C. */
#define G_INTVAL(x) intval(x)
#define G_SYMVAL(x) symval(x)
#define G_FUNCTORP(x) functorp(x)

#define G_functor(fa)  CONCATENATE(functor_, fa)
#define G_atom(fa)     CONCATENATE(atom_, fa)
#define G_stringify(s) STRINGIFY(s)

#define G_rappend0(class,post) CONCATENATE_3(class, _g_, post)
#define G_guard0(class,fa)     CONCATENATE_3(class, _g_guard_, fa)
#define G_body0(class,fa)      CONCATENATE_3(class, _g_body_, fa)

#define G_rappend(pref)  G_rappend0(CLASS_NAME, pref)

#define G_method_table0(class)  G_rappend0(class, method_table)

#define G_method_table  G_method_table0(CLASS_NAME)

#define G_CLASS_NAME_STRING  G_stringify(CLASS_NAME)

#define	G_ERROR_IN_NEW(errmsg) \
  G_error((errmsg), "creation", "", G_CLASS_NAME_STRING)

#define G_RETURN_FROM_PRINT()  do{ return 1L; }while(0)

#define G_RETURN_FROM_GC(newgobj)  do{ return (q*) (newgobj); }while(0)

#define G_UNIFY_VALUE(x, y) \
do{ \
  if( isref(x) && derefone(x) == (x) ){ \
    derefone(x) = (y); \
  }else{ \
    do_shallow_unify_value((x), (y)); \
  } \
}while(0)

#define G_SWITCH_ON_TERM(cons,atomic,funct,dobj,susp) \
do{ \
  for(;;){ \
    q _temp; \
    switch( ptagof(g_term) ){ \
    case ATOMIC: goto atomic; \
    case CONS:   goto cons; \
    case FUNCTOR: \
      if( isref(functor_of(g_term)) ) goto dobj; \
      goto funct; \
    case VARREF: \
      _temp = derefone(g_term); \
      if( isref(_temp) && g_term == derefone(_temp) ) goto susp; \
      g_term = _temp; \
      continue; \
    } \
  } \
}while(0)

#define G_DEF_NEW()  extern q G_rappend(new) (long g_argc, q* g_argv)

#define G_STD_DECL_FOR_NEW() \
  G_STD_DECL; \
  q (*g_myself)() = G_rappend(new)

#define G_DEREF_FOR_NEW(x) \
do{ \
  while( isref(x) ){ \
    q _temp = derefone(x); \
    if( isref(_temp) && (x) == derefone(_temp) ){ \
      q _var; \
      struct goalrec* _goal; \
      G_MAKE_THE_NEW_GOAL(&_var, &_goal, g_myself, g_argc, g_argv); \
      G_SUSPEND((x), _goal); \
      return _var; \
    }else{ \
      (x) = _temp; \
    } \
  } \
}while(0)


/** G_HEAPALLOC: procedure
 * from: type&
 * size > 0: offset_t
 * type: type cast or function returning q
 */
#define G_HEAPALLOC(from,size,type) \
do{ \
  (from) = type(klic_alloc(size)); \
}while(0)


/** G_HEAPALLOC_WITH_CHECK: procedure
 * from: type&
 * size > 0: offset_t
 * type: type cast or function returning q
 * res: {GENERIC_GCREQUEST, GENERIC_SUCCEEDED}&
 */
#define G_HEAPALLOC_WITH_CHECK(from,size,type,res) \
do{ \
  if ((long) real_heaplimit() <= (long) (heapp() + (size))) { \
    add_this_more_space(size); \
    heaplimit = 0; \
    (res) = GENERIC_GCREQUEST; \
  } else { \
    G_HEAPALLOC((from), (size), type); \
    (res) = GENERIC_SUCCEEDED; \
  } \
}while(0)


/** G_MAKE_VAR: procedure
 * x: q&
 */
#define G_MAKE_VAR(x) \
do{ \
  G_HEAPALLOC((x), 1, (q)); \
  derefone(x) = (x); \
}while(0)


#define G_STD_DECL struct global_variables* glbl = My_glbl

#define G_SIZE_IN_Q(type) ((sizeof(type)+sizeof(q)-1) / sizeof(q))

/* runtime/gc.c */
extern void push_gc_stack(q* addr);


/** G_COPY_KL1_TERM_TO_NEWGEN: procedure
 * from: q
 * to: q&
 */
#define G_COPY_KL1_TERM_TO_NEWGEN(from,to) \
do{ \
  if(isref(from)) { \
    q* newword = klic_alloc(1); \
    derefone(newword) = (from); \
    (to) = makeref(newword); \
    push_gc_stack(newword); \
  } else { \
    (to) = (from); \
    push_gc_stack(&(to)); \
  } \
}while(0)


/** G_PUSH_GOAL: procedure
 * goal: struct goalrec*
 */

#define G_PUSH_GOAL(goal) \
do{ resume_same_prio(goal); }while(0)

#define G_PUSH_GOAL_TO_SPECIAL_QUEUE(goal) \
  G_PUSH_GOAL(goal)


/* deta flag for Object Table */
enum g_type { G_DATA, G_CONSUMER, G_GENERATOR };


/* external functions */


/* runtime/generic.c */

extern NeverReturn void
G_error(char* errmsg, char* where, char* object_kind, char* class);

extern void
G_MAKE_THE_NEW_GOAL(q* var, struct goalrec** goalp1,
  q (*myself)(), unsigned long argc, q* argv );


#endif /* _KLIC_G_BASIC_H_ */
